// Lawson Labs, Inc.
// 3239 Phoenixville Pike
// Malvern, PA  19355
// 800 321-5355 or 610 725-8800
// August 5, 1998
// Tim Van Dusen - programmer
// 32-bit Windows-9x console application

#define STRICT
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <dos.h>
#ifdef __cplusplus
	 #define __CPPARGS ...
#else
	 #define __CPPARGS
#endif
#include "type_ide.h"
#include "m141_95.h"

#ifdef MS_COMPILER
	#define __emit__ _asm _emit
#endif

//time in seconds for each tick (GetTickCount resolution)
const double tickTime = .001;
char tmpbuf[80];

void SetWin32CursorOn(void);
void SetWin32CursorOff(void);

HANDLE	hOutCon, hInCon;
DWORD	lpNumberOfCharsWritten;
COORD	pos, dwCursorPosition;
CONSOLE_CURSOR_INFO	ConsoleCursorInfo;
CONSOLE_SCREEN_BUFFER_INFO lpConsoleScreenBufferInfo;




int main(void)
{
	char	ch;
	short calFlag = 0;
	int n, p_row, p_col, incFlag, endDisplay, cStat, sdCount=1, sdGrp=40;
	long i, j;
	unsigned long arraySize;
	unsigned long *StoreConv;
	double scanTime, maxHertzRate, voltage, max=-6000, min=6000;
	double stDv=0, rmsBits=0, sdTot = 0, runningAvg=0, sd[201];
	DWORD tckTime;
	HANDLE hglbStoreConv;
	T_M141 M141;

	hglbStoreConv = NULL;

	hOutCon = GetStdHandle(STD_OUTPUT_HANDLE); 
	hInCon =  GetStdHandle(STD_INPUT_HANDLE);
	
	if(!hOutCon) 
	{
		printf("Failed to get a handle to STDOUT!");
		printf("\nany key to exit");
		getch();
		exit(0);
	}
 
	pos.X = 0;
	pos.Y = 0;
	SetConsoleCursorPosition(hOutCon, pos);
	ConsoleCursorInfo.dwSize=1;
	ConsoleCursorInfo.bVisible = TRUE;

	if (!ReadConfig(&M141))   
	{
		pos.X = 24;	
		pos.Y = 10;	 
		SetConsoleCursorPosition(hOutCon, pos);  
		printf("unable to read 141EWIN.CFG file");
		pos.X = 26;	
		pos.Y = 11;	 
		SetConsoleCursorPosition(hOutCon, pos);  
		printf("press any key to continue");
		getch();
	}

TOP_MENU:
	//if handle to storage array exists, free it
	if (hglbStoreConv != NULL)
	{
		 GlobalUnlock(hglbStoreConv);
		 GlobalFree(hglbStoreConv);
	}

	ch = 0;  


	while(ch != 27)
	{
		
		pos.X = 0;	
		pos.Y = 0;	
		FillConsoleOutputCharacter(hOutCon, ' ',	
			80*25, pos,	&lpNumberOfCharsWritten);	
		pos.X = 27;	
		pos.Y = 4 ;	 
		SetConsoleCursorPosition(hOutCon, pos);		 
		printf ("MODEL 141");
		pos.X = 27;	
		pos.Y = 6;	 
		SetConsoleCursorPosition(hOutCon, pos);		 
		printf ("Main Menu");
		pos.X = 20;	
		pos.Y = 8;	 
		SetConsoleCursorPosition(hOutCon, pos);		 
		printf ("1) ADC              %u",M141.adc141);
		pos.X = 20;	
		pos.Y = 10;	 
		SetConsoleCursorPosition(hOutCon, pos);		 
		printf ("2) Polled Mode Operation");
		pos.X = 20;	
		pos.Y = 12;	 
		SetConsoleCursorPosition(hOutCon, pos);		 
		printf ("3) Scanning Mode Operation");
		pos.X = 18;	
		pos.Y = 14;	 
		SetConsoleCursorPosition(hOutCon, pos);		 
		printf ("Esc) Exit program");

		ch = getch();

		if (ch == 27)
		{
			goto EXIT_MAIN;
		}

		if (ch == '1')
		{
			//get and check ADC address
			calFlag = GetADC(&M141);
		}

		if (ch == '2')
		{
			//if not calibrated, ADC address from config file not verified
			if (calFlag == 0)
			{

				//check for 141 at ADC address
				if (ADInit(M141.adc141)==0)
				{
					
					pos.X = 0;	
					pos.Y = 0;	
					FillConsoleOutputCharacter(hOutCon, ' ',	
						80*25, pos,	&lpNumberOfCharsWritten);	
					SetConsoleCursorPosition(hOutCon, pos);		 
					printf ("\nModel 141 not found at ADC %u\n",M141.adc141);
					printf("press any key to continue");
					getch();
					goto TOP_MENU;
				}

				pos.X = 1;	
				pos.Y = 24;	 
				SetConsoleCursorPosition(hOutCon, pos);			 
				printf("                           Calibrating...                       ");
				if (Calibrate(&M141)==0)
				{
					
					pos.X = 0;	
					pos.Y = 0;	
					FillConsoleOutputCharacter(hOutCon, ' ',	
						80*25, pos,	&lpNumberOfCharsWritten);	
					SetConsoleCursorPosition(hOutCon, pos);		 
					printf ("\nCalibration was not successful\n");
					printf("press any key to continue");
					getch();
					goto TOP_MENU;
				}
				calFlag = 1;
			}

			//initialize standard deviation count
			sdCount=1;
			goto POLLED;
		}

		if (ch == '3')
		{
			//if not calibrated, ADC address from config file not verified
			if (calFlag == 0)
			{
				//check for 141 at ADC address
				if (ADInit(M141.adc141)==0)
				{
					
					pos.X = 0;	
					pos.Y = 0;	
					FillConsoleOutputCharacter(hOutCon, ' ',	
						80*25, pos,	&lpNumberOfCharsWritten);	
					SetConsoleCursorPosition(hOutCon, pos);		 
					printf ("\nModel 141 not found at ADC %u\n",M141.adc141);
					printf("press any key to continue");
					getch();
					goto TOP_MENU;
				}

				pos.X = 1;	
				pos.Y = 24;	 
				SetConsoleCursorPosition(hOutCon, pos);			 
				printf("                           Calibrating...                       ");
				if (Calibrate(&M141)==0)
				{
					
					pos.X = 0;	
					pos.Y = 0;	
					FillConsoleOutputCharacter(hOutCon, ' ',	
						80*25, pos,	&lpNumberOfCharsWritten);	
					SetConsoleCursorPosition(hOutCon, pos);		 
					printf ("\nCalibration was not successful\n");
					printf("press any key to continue");
					getch();
					goto TOP_MENU;
				}
				calFlag = 1;
			}

		goto GET_SCANS;
		}
	} //while Ch != 27

	goto EXIT_MAIN;

//polled mode example
POLLED:
	WaitSettle(&M141);

	ch = 0;
	p_row = 13;

	
	pos.X = 0;	
	pos.Y = 0;	
	FillConsoleOutputCharacter(hOutCon, ' ',					
			80*25, pos,	&lpNumberOfCharsWritten);				
	SetConsoleCursorPosition(hOutCon, pos);						 
	PrintCommandMenu(&M141);		

	//repeat until esc key pressed or an error
	while (ch != 27)
	{
		//loop until keyboard hit
		while (!kbhit())
		{
			SetWin32CursorOff();
			voltage = GetConv(&M141)* 1000;

			pos.X = 2;			
			pos.Y = p_row++;	 
			SetConsoleCursorPosition(hOutCon, pos);				
			SetWin32CursorOff();								

			printf ("  %10.3lf      ",voltage);

			if (max < voltage)
			{
				//update minimum
				max = voltage;
			}

			if (min > voltage)
			{
				//update minimum
				min = voltage;
			}


			if (M141.calTime >= 5000)
			{
				sdGrp = 40;
			}
			else if ((M141.calTime > 1500) && (M141.calTime < 5000))
			{
				sdGrp = 100;
			}
			else if (M141.calTime <= 1500)
			{
				sdGrp = 200;
			}

			StandardDev(&stDv, &rmsBits, &sdTot, &runningAvg, sd,
							&voltage, &sdCount, M141.biPolar, sdGrp);

			pos.X = 57;	
			pos.Y = 5;	 
			SetConsoleCursorPosition(hOutCon, pos);				
			SetWin32CursorOff();								
			if (M141.digInp == 1)
			{
				printf("on ");
			}
			else
			{
				printf("off");
			}

			if (p_row == 23)
			{
			p_row = 13;

				pos.X = 40;	
				pos.Y = 14;	 
				SetConsoleCursorPosition(hOutCon, pos);			
				SetWin32CursorOff();							
				printf ("Maximum     %8.3lf mv       ", max);
				pos.X = 40;	
				pos.Y = 15;	 
				SetConsoleCursorPosition(hOutCon, pos);			
				SetWin32CursorOff();							
				printf ("Minimum     %8.3lf mv       ", min);
				pos.X = 40;	
				pos.Y = 16;	 
				SetConsoleCursorPosition(hOutCon, pos);			
				SetWin32CursorOff();							
				printf ("Delta       %8.3lf mv       ", fabs(max - min));
				pos.X = 40;	
				pos.Y = 17;	 
				SetConsoleCursorPosition(hOutCon, pos);			
				SetWin32CursorOff();							
				printf ("Average     %8.3lf mv       ", runningAvg);
				pos.X = 40;	
				pos.Y = 19;	 
				SetConsoleCursorPosition(hOutCon, pos);			
				SetWin32CursorOff();							
				printf ("Standard deviation  %8.3lf mv   ", stDv);
				pos.X = 40;	
				pos.Y = 20;	 
				SetConsoleCursorPosition(hOutCon, pos);			
				SetWin32CursorOff();							
				printf ("Effective Resolution  %5.1lf  bits", rmsBits);
			}
		}   //while (!kbhit())

		ch = getch();

		if (ch != 27)
		{

			if (islower(ch))
			{
				ch = toupper(ch);
			}

			SetWin32CursorOn();
			DoCommand (&M141, ch, &max, &min);
		}

	}//while (Ch != 27)

	 //end of polled mode example
	goto TOP_MENU;

//scanning mode
GET_SCANS:

	//if handle to storage array exists,
	if (hglbStoreConv != NULL)
	{
		 //free handle
		 GlobalUnlock(hglbStoreConv);
		 GlobalFree(hglbStoreConv);
	}

	if (!SetupScan(&M141, &maxHertzRate))
	{
	  goto TOP_MENU;
	}

GET_NUM_SCANS:
	
	pos.X = 0;	
	pos.Y = 0;	
	FillConsoleOutputCharacter(hOutCon, ' ',				
			80*25, pos,	&lpNumberOfCharsWritten);			
	SetConsoleCursorPosition(hOutCon, pos);					 
	printf("%lu scans per sampling run (Y/N)? ",M141.totScans);
	ch = getch();

	if (ch == 27)
	{
		goto TOP_MENU;
	}

	if (ch == 'N' || ch == 'n')
	{
		printf("\n\nEnter scans per sampling run ");
		getch();
		fgets(tmpbuf,80,stdin);
		sscanf(tmpbuf,"%lu",&M141.totScans);

		if (M141.totScans == 0)
		{
			M141.totScans = 1;
		}

		goto GET_NUM_SCANS;
	}
	else if (ch != 'Y' && ch != 'y')
	{
		goto GET_NUM_SCANS;
	}

	scanTime = (1/M141.hertzRate)*(M141.totScans-1) + 1.0/maxHertzRate;

	//each datapoint is 4 bytes
	arraySize = (M141.totScans*M141.numData)*4;

	hglbStoreConv=GlobalAlloc(GPTR,arraySize);
	if (hglbStoreConv == NULL)
	{
		 printf("\n\nunable to allocate %ld bytes. Reduce number of scans\n",arraySize);
		 printf("any key to continue");  
		 getch();
		 goto GET_NUM_SCANS;
	}

	//allocate storage area
	StoreConv = (unsigned long  *)GlobalLock(hglbStoreConv);

DISPLAY:
	
	pos.X = 0;	
	pos.Y = 0;	
	FillConsoleOutputCharacter(hOutCon, ' ',					
			80*25, pos,	&lpNumberOfCharsWritten);				
	SetConsoleCursorPosition(hOutCon, pos);						 

	printf("\nData points per scan = %u\n", M141.numData);
	printf("Average %u samples per data point\n",M141.average);
	printf("Inter-channel delay = %u  milliseconds\n",M141.ichDelay);
	printf("Scans per sample run = %lu\n", M141.totScans);
	printf("Size of array = %lu bytes\n",arraySize);
	printf("Scanning rate = %8.2lf hz\n",M141.hertzRate);
	printf("Sample run duration %8.2lf seconds\n",scanTime);
	printf("\n\nIs everything correct (Y/N)? ");
	ch = getch();

	if (ch == 27)
	{
	  goto TOP_MENU;
	}

	if (ch == 'N' || ch == 'n')
	{
	  goto GET_SCANS;
	}
	else if (ch != 'Y' && ch != 'y')
	{
	  goto DISPLAY;
	}

REVIEW_CODES:
	
	pos.X = 0;	
	pos.Y = 0;	
	FillConsoleOutputCharacter(hOutCon, ' ',					
			80*25, pos,	&lpNumberOfCharsWritten);				
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf ("Review or revise control codes (Y/N)? ");
	ch = getch();


	if (ch == 27)
	{
	  goto TOP_MENU;
	}

REVIEW_CODES2:

	if (ch == 'Y' || ch == 'y')
	{
		DispCodes(&M141);

		//initialize starting row and col
		p_row = 4;
		p_col = 1;

		//not incrementing
		incFlag = 0;

		//initialize cCode[0] for increment
		M141.cCode[0] = -1;

		//for (n = 1;n <= M141.numData;n++)
		for (n = 1;n <= (int)M141.numData;n++)					
		{
			pos.X = p_col;										
			pos.Y = p_row++;									 
			SetConsoleCursorPosition(hOutCon, pos);				 

			//if not incrementing
			if (incFlag == 0)
			{
				if (M141.cCode[n] < 10)
				{
					pos.X = p_col+12;	
					pos.Y = p_row-1;	 
					SetConsoleCursorPosition(hOutCon, pos);		 
				}
				else
				{
					pos.X = p_col+11;	
					pos.Y = p_row-1;	 
					SetConsoleCursorPosition(hOutCon, pos);		 
				}

				//get control code for current channel
				getch();
				fgets(tmpbuf,80,stdin);

				//if entered 'i' then start incrementing
				if (tmpbuf[0] == 'i' || tmpbuf[0] == 'I')
				{
					incFlag = 1;

					//increment control code
					M141.cCode[n] = M141.cCode[n-1]+1;
					if (M141.cCode[n] >= 32)
					{
						M141.cCode[n] = 0;
					}
					pos.X = p_col+11;	
					pos.Y = p_row-1;	 
					SetConsoleCursorPosition(hOutCon, pos);		 
					printf ("%2u  ",M141.cCode[n]);
				}
				else
				{
					pos.X = p_col+12;	
					pos.Y = p_row-1;	 
					SetConsoleCursorPosition(hOutCon, pos);		 
					printf(" ");
					pos.X = p_col+12;	
					pos.Y = p_row-1;	 
					SetConsoleCursorPosition(hOutCon, pos);		 

					//store current control code
					sscanf(tmpbuf,"%u",&M141.cCode[n]);

					//control codes must be >= 0 and < 32
					if (M141.cCode[n] >= 32)
					{
						M141.cCode[n] = 31;
					}
				}
			} //if incFlag == 0
			else if (incFlag == 1)
			{
				M141.cCode[n] = M141.cCode[n-1]+1;

				//control code rolls over at 32
				if (M141.cCode[n] >= 32)
				{
				  M141.cCode[n] = 0;
				}
			}

			pos.X = p_col+11;	
			pos.Y = p_row-1;	 
			SetConsoleCursorPosition(hOutCon, pos);				 

			//keep display in window
			if ((p_col + 20) > 60)
			{
				printf ("%2u    ",M141.cCode[n]);
			}
			else
			{
				printf ("%2u       ",M141.cCode[n]);
			}

			//keep display in window
			if (p_row == 20 && (n != (int)M141.numData))
			{
				p_col += 19;
				p_row = 4;
			}

		} //end for loop

		M141.cCode[n] = M141.cCode[1];

EXIT_CODES:
		pos.X = 10;	
		pos.Y = 24;	 
		SetConsoleCursorPosition(hOutCon, pos);					 
		printf ("Revise control codes (Y/N)?                                     ");
		pos.X = 38;	
		pos.Y = 24;	 
		SetConsoleCursorPosition(hOutCon, pos);					 
		ch = getch();

		if (ch == 27)
		{
			goto TOP_MENU;
		}

		if (ch == 'Y' || ch == 'y')
		{
			goto REVIEW_CODES2;
		}
		else if (ch != 'N' && ch != 'n')
		{
			goto EXIT_CODES;
		}

	}
	else if (ch != 'N' && ch != 'n')
	{
	  goto REVIEW_CODES;
	}

	//select first channel
	funcOutp(M141.adc141, M141.cCode[1]+M141.biPolar);
	WaitSettle(&M141);

	
	pos.X = 0;	
	pos.Y = 0;	
	FillConsoleOutputCharacter(hOutCon, ' ',					
			80*25, pos,	&lpNumberOfCharsWritten);				
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf("\nSampling will take %8.2lf seconds",scanTime);
	printf("\npress a key to start scanning, ESC to exit");

	if (getch() == 27)
	{
	  goto TOP_MENU;
	}

	pos.X = 1;	
	pos.Y = 3;	 
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf("Activating digital input will cut scan short\n");
	pos.X = 1;	
	pos.Y = 4;	 
	SetConsoleCursorPosition(hOutCon, pos);						 

	//get start time in milliseconds
	tckTime = GetTickCount();
	cStat = Call141(&M141, StoreConv);

	//get elapsed time in milliseconds
	tckTime = GetTickCount() - tckTime;
	if (cStat == 0)
	{
      printf ("\nUnable to find Model 141\n");
	  printf("press any key to continue");
	  getch();
	  goto TOP_MENU;
	}
	else if(cStat == 2)
	{
	  printf ("Sampling stopped. ");
	}

	printf ("%lu of %lu scans completed\n",M141.numCompleted, M141.totScans);

	if(cStat == 1)
	{
	  printf ("\nSampling complete.\n");
	}
	printf ("Elapsed time:   %8.2lf",tckTime/1000.0);

SAVE_FILE:
	pos.X = 1;	
	pos.Y = 8;	 
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf("Do you want to save the data as a text file (Y/N)?     ");
	pos.X = 52;	
	pos.Y = 8;	 
	SetConsoleCursorPosition(hOutCon, pos);						 
	ch = getch();

	if (ch == 'Y' || ch == 'y')
	{
		//save data to disk
		SaveData(&M141, StoreConv);
	}
	else if (ch != 'N' && ch != 'n')
	{
		goto SAVE_FILE;
	}


// start display of scans

	//formatting to keep display in window
	endDisplay =  (int)(ceil((double)(M141.numData)/8.0));		
	
	pos.X = 0;	
	pos.Y = 0;	
	FillConsoleOutputCharacter(hOutCon, ' ',					
			80*25, pos,	&lpNumberOfCharsWritten);				
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf("any key to continue, ESC to end                    ");
	pos.X = 0;	
	pos.Y = 1;	 
	SetConsoleCursorPosition(hOutCon, pos);						 

	for (i = 0; i < (int)M141.numCompleted; i++)				
	{
	  printf("%-4ld", i+1);

	  for (j = 0; j < (int)M141.numData; j++)
	  {
		 //calculate voltage for bipolar or unipolar
		 if (M141.biPolar)
		 {
			voltage = ((double)StoreConv[i*M141.numData + j])*9.536743e-3 - 5000;
		 }
		 else
		 {
			voltage = ((double)StoreConv[i*M141.numData + j])*4.768372e-3;
		 }

		 printf (" %8.3lf ",voltage);

		 //keep display in window
	     GetConsoleScreenBufferInfo(hOutCon, &lpConsoleScreenBufferInfo);
	     if(lpConsoleScreenBufferInfo.dwCursorPosition.X > 70)
		 {
			printf("\n    ");
		 }
	  }

	  printf ("\n");

	  //pause for user input
	  GetConsoleScreenBufferInfo(hOutCon, &lpConsoleScreenBufferInfo);
	  if(lpConsoleScreenBufferInfo.dwCursorPosition.Y > 24)
	  {
		 if (getch() == 27)
		 {
			goto END_SCANS;
		 }
		pos.X = 1;	
		pos.Y = 1;	 
		SetConsoleCursorPosition(hOutCon, pos);					 
	  }
	}

END_SCANS:
	pos.X = 1;	
	pos.Y = 25;	 
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf("R to retry, ESC for menu                              ");
	pos.X = 26;	
	pos.Y = 25;	 
	SetConsoleCursorPosition(hOutCon, pos);						 

RETRY:
	ch = getch();

	if (ch == 'R' || ch == 'r')
	{
	  goto DISPLAY;
	}
	else if( ch == 27)
	{
	  goto TOP_MENU;
	}

	goto RETRY;

EXIT_MAIN:

SAVE_CONFIG:
	
	pos.X = 0;	
	pos.Y = 0;	
	FillConsoleOutputCharacter(hOutCon, ' ',					
			80*25, pos,	&lpNumberOfCharsWritten);				
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf("\nDo you wish to save configuration (Y/N)? ");
	ch = getch();
	FillConsoleOutputCharacter(hOutCon, ' ',					
			80*25, pos,	&lpNumberOfCharsWritten);				
	pos.X = 0;	
	pos.Y = 0;	 
	SetConsoleCursorPosition(hOutCon, pos);						 

	if (ch == 'Y' || ch == 'y')
	{
		WriteConfig(&M141);
	}
	else if(ch != 'N' && ch != 'n')
	{
		goto SAVE_CONFIG;
	}

	//close window
	DestroyWindow(GetActiveWindow());
	exit(0);
	return 1;
}

//make window the same size as the desktop
void SetWindowToDesktop(void)
{
	RECT rect;

	
	pos.X = 0;	
	pos.Y = 0;	
	FillConsoleOutputCharacter(hOutCon, ' ',					
			80*25, pos,	&lpNumberOfCharsWritten);				
	SetConsoleCursorPosition(hOutCon, pos);						 
	GetWindowRect(GetDesktopWindow(), &rect);
	SetWindowPos(GetActiveWindow(), (HWND) NULL, 0, 0, rect.right,
					rect.bottom, SWP_NOZORDER | SWP_NOACTIVATE);
}

//get and check ADC address
short GetADC(T_M141  *M141)
{

GET_ADC:
	
	pos.X = 0;	
	pos.Y = 0;	
	FillConsoleOutputCharacter(hOutCon, ' ',					
			80*25, pos,	&lpNumberOfCharsWritten);				
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf("Enter ADC (check your manual) ");
	getch();
	fgets(tmpbuf,80,stdin);

	sscanf(tmpbuf,"%u",&M141->adc141);

	if (!CheckADC(M141->adc141))
	{
		 printf ("\n%u is not a valid ADC address \n",M141->adc141);
		 printf("press ENTER to try again");

		 if (getch() == 27)
		 {
			return 0;
		 }
		 goto GET_ADC;
	}

	if (ADInit(M141->adc141)==0)
	{
		 printf ("\nModel 141 not found at ADC %u\n",M141->adc141);
		 printf("press ENTER to try again, ESC to abort");

		 if (getch() == 27)
		 {
			return 0;
		 }

		 goto GET_ADC;
	}

	pos.X = 1;	
	pos.Y = 24;	 
	SetConsoleCursorPosition(hOutCon, pos);  
	printf("                           Calibrating...                       ");
	if (Calibrate(M141)==0)
	{
		// clrscr();
		pos.X = 0;	
		pos.Y = 0;	
		FillConsoleOutputCharacter(hOutCon, ' ',				
			80*25, pos,	&lpNumberOfCharsWritten);				
		 printf ("\nCalibration was not successful\n");
		 printf("press any key to continue");
		 getch();
		 return 0;
	}
	return 1;
}

//print polled mode command menu
void PrintCommandMenu(T_M141  *M141)
{
	pos.X = 15;	
	pos.Y = 1;	 
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf ("POLLED  MENU   Lawson Labs Model 141");
	pos.X = 5;	
	pos.Y = 3;	 
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf ("C) A/D input Channel    %u", (M141->controlCode>>4));
	pos.X = 5;	
	pos.Y = 4;	 
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf ("E) External sub-channel %u",M141->controlCode&0xF);

	if ((M141->controlCode&0xF) < 10)
	{
		printf (" ");
	}

	pos.X = 5;	
	pos.Y = 5;	 
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf ("P) ");

	if (M141->biPolar == 0x80)
	{
	 printf("biPolar ");
	}
	else
	{
	 printf("uniPolar");
	}

	pos.X = 40;	
	pos.Y = 22;	 
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf ("A/D clock ");
	if (M141->calTime >= 5000)
	{
		printf("224 KHz  ");
	}
	else if ((M141->calTime > 1500) && (M141->calTime < 5000))
	{
		printf("895 KHz  ");
	}
	else if (M141->calTime <= 1500)
	{
		printf("3.58 MHz  ");
	}

	pos.X = 43;	
	pos.Y = 3;	 
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf ("A) Averaging            %-5u", M141->average);

	pos.X = 43;	
	pos.Y = 4;	 
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf ("R) Recalibrate");

	pos.X = 43;	
	pos.Y = 5;	 
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf ("Digital input ");
	if (M141->digInp == 1)
	{
		printf("on ");
	}
	else
	{
		printf("off");
	}

	pos.X = 6;	
	pos.Y = 11;	 
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf ("Millivolts\n");
	pos.X = 1;	
	pos.Y = 24;	 
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf (" Press ESC for menu                                              ");
}

void DoCommand(T_M141  *M141, char ch, double *max, double *min)
{
	int     tempExt;

	if (ch == 'A')
	{
		//get averaging
		M141->average = GetAvg(M141->average);		
	}

	//toggle channel
	if (ch == 'C')
	{
		if (M141->controlCode > 15)
		{
			M141->controlCode = M141->controlCode&0xF;
		}
		else
		{
			M141->controlCode = M141->controlCode + 0x10;
		}

		//select current channel
		funcOutp(M141->adc141, M141->controlCode+M141->biPolar);
		 WaitSettle(M141);
	}

	if (ch == 'E')
	{
		tempExt = -1;
		pos.X = 1;	
		pos.Y = 24;	 
		SetConsoleCursorPosition(hOutCon, pos);					 
		printf("                                    ");

		while ((tempExt < 0) || (tempExt > 15))
		{
			pos.X = 1;	
			pos.Y = 24;	 
			SetConsoleCursorPosition(hOutCon, pos);				 
			printf("Enter sub channel ");
			pos.X = 19;	
			pos.Y = 24;	 
			SetConsoleCursorPosition(hOutCon, pos);				 
			getch();
			fgets(tmpbuf,80,stdin);
			sscanf(tmpbuf,"%d",&tempExt);
		}

		M141->controlCode = ((M141->controlCode>>4)<<4) + tempExt;

		//select current channel
		funcOutp(M141->adc141, M141->controlCode+M141->biPolar);

		WaitSettle(M141);
	}

	if (ch == ' ')
	{
		 //reset max-min
		 (*max) = -6000.0;
		 (*min) = 6000.0;
	}

	if (ch == 'R')
	{
		pos.X = 1;	
		pos.Y = 24;	 
		SetConsoleCursorPosition(hOutCon, pos);					 
		printf("                           Calibrating...                       ");
		if (Calibrate(M141)==0)
		{
			
			pos.X = 0;	
			pos.Y = 0;	
			FillConsoleOutputCharacter(hOutCon, ' ',			
				80*25, pos,	&lpNumberOfCharsWritten);			
			SetConsoleCursorPosition(hOutCon, pos);				 
			printf ("\nCalibration was not successful\n");
			printf("press any key to continue");
			getch();
		}
		WaitSettle(M141);
	}

	//toggle polarity
	if (ch == 'P')
	{

		if (M141->biPolar == 0)
		{
			M141->biPolar = 0x80;
		}
		else
		{
			M141->biPolar = 0;
		}

		pos.X = 1;	
		pos.Y = 24;	 
		SetConsoleCursorPosition(hOutCon, pos);					 
		printf("                           Calibrating...                       ");
		if (Calibrate(M141)==0)
		{
			
			pos.X = 0;	
			pos.Y = 0;	
			FillConsoleOutputCharacter(hOutCon, ' ',			
				80*25, pos,	&lpNumberOfCharsWritten);			
			SetConsoleCursorPosition(hOutCon, pos);				 
			printf ("\nCalibration was not successful\n");
			printf("press any key to continue");
			getch();
		}

		//select current channel
		funcOutp(M141->adc141, M141->controlCode+M141->biPolar);
		WaitSettle(M141);
	}

	PrintCommandMenu(M141);

}

//get averaging (1,2,4,8...256)
unsigned int GetAvg(unsigned int average)
{
	int tempCh = 0;

	while((tempCh != 27) && (tempCh != 13))
	{
		while (!kbhit())
		{
			pos.X = 1;	
			pos.Y = 24;	 
			SetConsoleCursorPosition(hOutCon, pos);				 
			printf ( "Press + increase, - decrease averaging, Enter to continue ");
			printf ( "%-6u", average);
		}

		tempCh =(short) getch();

		if (tempCh == '+')
		{
			average *= 2;
			if (average == 0)
			{
			  average = 1;
			}
		}

		if (tempCh == '-')
		{
			average /= 2;
		}

		if (average  < 1)
		{
			average = 256;
		}
		else if (average > 256)
		{
			average = 1;
		}
	}
	return average;
}

void SaveData(T_M141  *M141, unsigned long  *dataBuffer)
{
	char textFile[13];
	long i, j;
	double voltage;

	//handle for saving text file to disk
	FILE *tFile;

	printf("\n\nEnter the name of the text file ");
	getch();
	fgets(tmpbuf,80,stdin);

	sscanf(tmpbuf,"%13s",&textFile);

	if(strlen(tmpbuf))
	{
		if ((tFile = fopen(textFile, "wt")) != NULL)
		{
			printf("Saving data...");
			for (i = 0; i < (int)M141->numCompleted; i++)
			{
				fprintf(tFile,"%-5ld ", i+1);
				for (j = 0; j < (int)M141->numData; j++)
				{
					if (M141->biPolar)
					{
						voltage = ((double)dataBuffer[i*M141->numData + j])
									*9.536743e-3 - 5000;
					}
					else
					{
						voltage = ((double)dataBuffer[i*M141->numData + j])
									*4.768372e-3;
					}
					fprintf (tFile," %8.3lf",voltage);
				}
				fprintf (tFile,"\n");
			}
			fclose(tFile);
			pos.X = 1;	
			GetConsoleScreenBufferInfo(hOutCon,					
				&lpConsoleScreenBufferInfo);					
			pos.Y = lpConsoleScreenBufferInfo.dwCursorPosition.Y;  
			SetConsoleCursorPosition(hOutCon, pos);				 
			printf("Data saved in file %s",textFile);
	 }
	 else
	 {
		printf("unable to open file %s\nPress any key to continue",textFile);
		getch();
	 }
  }
}

//display control codes
void DispCodes(T_M141  *M141)
{
	int p_row, p_col, n;

	
	pos.X = 0;	
	pos.Y = 0;	
	FillConsoleOutputCharacter(hOutCon, ' ',					
			80*25, pos,	&lpNumberOfCharsWritten);				
	pos.X = 1;	
	pos.Y = 1;	 
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf("  data\n");
	printf("  point   CCode\n");
	pos.X = 10;	
	pos.Y = 24;	 
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf("Press RETURN to leave unchanged");

	if (M141->numData > 1)
	{
	 printf(", I to increment ");
	}

	p_row = 4;
	p_col = 1;
	for (n = 1;n <= (int)M141->numData;n++)
	{
		pos.X = p_col;		
		pos.Y = p_row++;	 
		SetConsoleCursorPosition(hOutCon, pos);					 
		printf("   %2d      %2u  ", n, M141->cCode[n]);
		if (p_row == 20 && (n != (int)M141->numData))
		{
			p_col += 19;
			p_row = 4;
			pos.X = p_col;	
			pos.Y = 1;		 
			SetConsoleCursorPosition(hOutCon, pos);				 
			printf("  data");
			pos.X = p_col;	
			pos.Y = 2;		 
			SetConsoleCursorPosition(hOutCon, pos);				 
			printf("  point   CCode");
		}
	}
	M141->cCode[n] = M141->cCode[1];
}

short ReadConfig(T_M141  *M141)
{
	int n;
	FILE *rCfg;

	if ((rCfg = fopen("141ewin.cfg", "rt")) != NULL)
	{
		fgets(tmpbuf,80,rCfg);
		sscanf(tmpbuf,"%u",&M141->adc141);
		fgets(tmpbuf,80,rCfg);
		sscanf(tmpbuf,"%lf",&M141->hertzRate);
		fgets(tmpbuf,80,rCfg);
		sscanf(tmpbuf,"%u",&M141->average);
		fgets(tmpbuf,80,rCfg);
		sscanf(tmpbuf,"%u",&M141->ichDelay);
		fgets(tmpbuf,80,rCfg);
		sscanf(tmpbuf,"%u",&M141->numData);
		fgets(tmpbuf,80,rCfg);
		sscanf(tmpbuf,"%lu",&M141->totScans);
		fgets(tmpbuf,80,rCfg);
		sscanf(tmpbuf,"%u",&M141->biPolar);
		fgets(tmpbuf,80,rCfg);
		sscanf(tmpbuf,"%u",&M141->controlCode);

		for (n = 0;n<= 33;n++)
		{
			fgets(tmpbuf,80,rCfg);
			sscanf(tmpbuf,"%u",&M141->cCode[n]);
		}

		fclose(rCfg);
		return 1;
	}
	else
	{
		M141->adc141 = 560;
		M141->hertzRate = 1000;

		//assume settling time for 874 Hz rate
		M141->ichDelay = 568;
		M141->average = 1;
		M141->numData = 1;
		M141->totScans = 100;
		M141->controlCode = 0;

		//default bipolar
		M141->biPolar = 0x80;
		for (n=0;n<=33;n++)
		{
			M141->cCode[n] = 0;
		}

		return 0;
	}
}


short WriteConfig(T_M141  *M141)
{
	int n;
	FILE *wCfg;

	if ((wCfg = fopen("141ewin.cfg", "wt")) != NULL)
	{
		fprintf(wCfg,"%u\n",M141->adc141);
		fprintf(wCfg,"%lf\n",M141->hertzRate);
		fprintf(wCfg,"%u\n",M141->average);
		fprintf(wCfg,"%u\n",M141->ichDelay);
		fprintf(wCfg,"%u\n",M141->numData);
		fprintf(wCfg,"%lu\n",M141->totScans);
		fprintf(wCfg,"%u\n",M141->biPolar);
		fprintf(wCfg,"%u\n",M141->controlCode);
		for (n = 0;n<= 33;n++)
		{
			fprintf(wCfg,"%u\n",M141->cCode[n]);
		}
		fclose(wCfg);
		return 1;
	}
	return 0;
}

short SetupScan(T_M141  *M141, double *maxHertzRate)
{
	char ch;
	double updateTime;

GET_POINTS:
	
	pos.X = 0;	
	pos.Y = 0;	
	FillConsoleOutputCharacter(hOutCon, ' ',					
			80*25, pos,	&lpNumberOfCharsWritten);				
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf("data points/scan= %u (Y/N)? ",M141->numData);
	ch = getch();

	if (ch == 27)
	{
		return 0;
	}

	if (ch == 'N' || ch == 'n')
	{
		printf("\n\nEnter number of data points per scan ");
		getch();
		fgets(tmpbuf,80,stdin);

		sscanf(tmpbuf,"%u",&M141->numData);

		if (M141->numData == 0)
		{
			M141->numData = 1;
		}
		else if (M141->numData > 32)
		{
			M141->numData = 32;
		}

		goto GET_POINTS;
	}
	else if (ch != 'Y' && ch != 'y')
	{
		goto GET_POINTS;
	}

GET_AVG:
	
	pos.X = 0;	
	pos.Y = 0;	
	FillConsoleOutputCharacter(hOutCon, ' ',					
			80*25, pos,	&lpNumberOfCharsWritten);				
	SetConsoleCursorPosition(hOutCon, pos);						 
	printf("average %u samples/data point (Y/N)? ",M141->average);
	ch = getch();

	if (ch == 27)
	{
	 return 0;
	}

	if (ch == 'N' || ch == 'n')
	{
	 M141->average = GetAvg(M141->average);
	 goto GET_AVG;
	}
	else if (ch != 'Y' && ch != 'y')
	{
		goto GET_AVG;
	}

	if (M141->numData == 1)
	{
		M141->ichDelay = 0;
	}
	else if (M141->numData > 1)
	{

GET_ICH:
		
		pos.X = 0;	
		pos.Y = 0;	
		FillConsoleOutputCharacter(hOutCon, ' ',				
			80*25, pos,	&lpNumberOfCharsWritten);				
		pos.X = 1;	
		pos.Y = 10;	 
		SetConsoleCursorPosition(hOutCon, pos);					 
		printf("minimum settling time = ");

		if (M141->calTime >= 5000)
		{
			printf("2270 milliseconds");
		}
		else if ((M141->calTime > 1500) && (M141->calTime < 5000))
		{
			printf("568 milliseconds");
		}
		else if (M141->calTime <= 1500)
		{
			printf("142 milliseconds");
		}

		pos.X = 1;	
		pos.Y = 2;	 
		SetConsoleCursorPosition(hOutCon, pos);					 
		printf("interchannel delay =  %u  milliseconds (Y/N)? ",M141->ichDelay);
		ch = getch();

		if (ch == 27)
		{
			return 0;
		}

		if (ch == 'N' || ch == 'n')
		{
			printf("\n\nEnter Interchannel delay count ");
			getch();
			fgets(tmpbuf,80,stdin);

			sscanf(tmpbuf,"%u",&M141->ichDelay);

			goto GET_ICH;
		}
		else if (ch != 'Y' && ch != 'y')
		{
			goto GET_ICH;
		}
	}

GET_RATE:
	
	pos.X = 0;	
	pos.Y = 0;	
	FillConsoleOutputCharacter(hOutCon, ' ',					
			80*25, pos,	&lpNumberOfCharsWritten);				
	SetConsoleCursorPosition(hOutCon, pos);						 

	if (M141->calTime >= 5000)
	{
		updateTime = 1.0/218.0;
	}
	else if ((M141->calTime > 1500) && (M141->calTime < 5000))
	{
		updateTime = 1.0/874.0;
	}
	else if (M141->calTime <= 1500)
	{
		updateTime = 1.0/3500.0;
	}

	if (M141->numData == 1)
	{
		*maxHertzRate = 1.0/(updateTime * M141->average * M141->numData);
	}
	else
	{
		*maxHertzRate = 1.0/((updateTime*M141->average+((double)M141->ichDelay/1000.0)
									+ .0003) * M141->numData);
	}

	if (*maxHertzRate > 1000)
	{
		*maxHertzRate = 1000;
	}

	if (*maxHertzRate < 0)
	{
		*maxHertzRate = 1;
	}

	if ((M141->hertzRate+.005) > *maxHertzRate)
	{
		printf("%8.2lf Max Rate (Y/N)? ",*maxHertzRate);
	}
	else
	{
		printf("%8.2lf scan rate (Y/N)? ",M141->hertzRate);
	}

	ch = getch();

	if (ch == 27)
	{
		return 0;
	}

	if (ch == 'N' || ch == 'n')
	{
		printf ("\n\nMax Rate %8.2lf", *maxHertzRate);
		printf("\n\nEnter scan rate ");
		getch();
		fgets(tmpbuf, 80, stdin);

		sscanf(tmpbuf, "%lf", &M141->hertzRate);

		if (M141->hertzRate <= 0)
		{
			M141->hertzRate = 1;
		}

		if (M141->hertzRate < 0)
		{
			M141->hertzRate = 1;
		}

		if (M141->hertzRate > *maxHertzRate)
		{
			M141->hertzRate = *maxHertzRate;
		}

		M141->hertzRate = 1.0/((((1.0/(double)M141->hertzRate)/ tickTime))*tickTime);

		goto GET_RATE;
	}
	else if (ch != 'Y' && ch != 'y')
	{
		goto GET_RATE;
	}

	if (M141->hertzRate > *maxHertzRate)
	{
		M141->hertzRate = *maxHertzRate;
	}

	return 1;
}

short CheckADC(unsigned int adc141)
{
	if (adc141 == 544 ||
		adc141 == 552 ||
		adc141 == 560 ||
		adc141 == 568 ||
		adc141 == 576 ||
		adc141 == 584 ||
		adc141 == 592 ||
		adc141 == 600 ||
		adc141 == 608 ||
		adc141 == 616 ||
		adc141 == 624)
	{
		return 1;
	}
	return 0;
}

//standard deviation and RMS resolution calculation
void    StandardDev(double *stDv, double *rmsBits, double *sdTot,
						  double *runningAvg, double *sd, double *voltage,
						  int *sdCount, int	biPolar, int sdGrp)
{
	int i = 0, nnn;
	double tempStDv, StDvRms;

	sd[*sdCount] = (*voltage);
	(*sdTot) += (*voltage);

	if ((*sdCount) >= sdGrp)
	{
		(*stDv) = 0; (*sdTot) = (*sdTot) / sdGrp;

		for (nnn = 1; nnn != sdGrp; nnn++)
		{
			(*stDv) = (*stDv) + pow(((*sdTot) - sd[nnn]), 2);
		}

		(*stDv) = sqrt((*stDv) / sdGrp);

		if (biPolar == 0x80)
		{
			StDvRms = (*stDv) / 9.536743e-3;
		}
		else
		{
			StDvRms = (*stDv) / 4.768372e-3;
		}

		tempStDv = StDvRms;

CALCNEXT:

		if (tempStDv >= 1)
		{
			i = i + 1;
			tempStDv = tempStDv / 2;
			goto CALCNEXT;
		}

		(*rmsBits) = 20 - i;

		if ((tempStDv < .75) && (i != 0))
		{
			(*rmsBits) = (*rmsBits) + .5;
		}

		(*runningAvg) = (*sdTot);
		(*sdTot) = 0; (*sdCount) = 1;
	}
	else
	{
		(*sdCount)++;
	}

}

//  If a new channel is selected or a calibration is
//  performed, wait time for channel voltage to settle
void WaitSettle(T_M141  *M141)
{
	DWORD settle, start;

	if (M141->calTime >= 5000)
	{
	 settle = 2270;
	}
	else if ((M141->calTime > 1500) && (M141->calTime < 5000))
	{
	 settle = 568;
	}
	else if (M141->calTime <= 1500)
	{
	 settle = 142;
	}

	pos.X = 1;	
	pos.Y = 24;	 
	SetConsoleCursorPosition(hOutCon, pos);  
	printf("                        waiting settling time...                        ");
	start = GetTickCount() + settle;
	while (start > GetTickCount()); 				//wait settling time
	pos.X = 1;	
	pos.Y = 24;	 
	SetConsoleCursorPosition(hOutCon, pos);  
	printf("                                              ");
}


//==================================================
// marker  -- PREVIOUSLY DLL (WIN-3.1 EASY-WIN TYPE)
//==================================================

char timerFlag=0;

//local prototypes
short ReadStatus(T_M141 *M141, unsigned int status);


short ADInit(ULONG ADC)
{
	DWORD start;

	//make sure not calibrating
	funcOutp(ADC, 128);

	//assert the FREEZE bit
	funcOutp(ADC+1, 1);

	if ((funcInp(ADC+3)) < 128)
	{
		return 0;
	}

	//now remove the FREEZE bit
	funcOutp(ADC+1, 0);

	start = GetTickCount()+1000;

	while ((start > GetTickCount()) && ((funcInp(ADC+3)) > 127));
	if ((funcInp(ADC+3)) > 127)
	{
	 return 0;
	}

	//start conversion if 140 board
	funcOutp(ADC+1,0);

	//must be a 140
	if (funcInp(ADC+3) > 127)
	{
	 return 0;
	}

	//card appears to be there
	return 1;
}


short Calibrate(T_M141  *M141)
{
	DWORD start;
	int n;

	start = GetTickCount();

	//freeze
	(funcOutp(M141->adc141+1, 1));

	//offset calibration
	(funcOutp(M141->adc141,48+M141->biPolar));

	//set calibration bit
	(funcOutp(M141->adc141,48+64+M141->biPolar));

	//clear calibration bit
	(funcOutp(M141->adc141,48+M141->biPolar));

	//unfreeze
	(funcOutp(M141->adc141+1, 0));

	// set up a machine dependent loop counter
	n = 0;
	while (n < 2000)
	{
		if (ReadStatus(M141, M141->adc141+3)!=0)
		{
			goto Offset_Good;
		}
		n++;
	}
	return 0;

Offset_Good:

	//freeze
	(funcOutp(M141->adc141+1, 1));

	//full scale calibration
	(funcOutp(M141->adc141,32+M141->biPolar));

	//set calibration bit
	(funcOutp(M141->adc141,32+64+M141->biPolar));

	//clear calibration bit
	(funcOutp(M141->adc141,32+M141->biPolar));

	//unfreeze
	(funcOutp(M141->adc141+1, 0));

	// set up a machine dependent loop counter
	n = 0;
	while (n < 2000)
	{
		if (ReadStatus(M141, M141->adc141+3)!=0)
		{
			goto FullScale_Good;
		}
		n++;
	}
	return 0;

FullScale_Good:


	//select current channel
	(funcOutp(M141->adc141, M141->controlCode+M141->biPolar));

	//get miliseconds elapsed
	M141->calTime = GetTickCount() - start;

	return 1;
}

/*
unsigned long  GetCount(T_M141  *M141)
{
	static unsigned long	tempavg, c1, c2, c3;
	static unsigned long avgConv, conversion;

	tempavg = 1;
	avgConv = 0;

	funcOutp(M141->adc141, M141->controlCode+M141->biPolar);		//select channel

Get_Conv_Again:
	if (ReadStatus(M141, M141->adc141+3)==0)
	{
	 goto Get_Conv_Error;
	}

	//assert FREEZE
	funcOutp(M141->adc141+1, 1);

	c3 = (funcInp(M141->adc141+2))&0xF;
	c2 = funcInp(M141->adc141+1);
	c1 = funcInp(M141->adc141);

	if ((funcInp(M141->adc141+3))&16)
	{
	 M141->digInp = 0;
	}
	else
	{
	 M141->digInp = 1;
	}

	//unfreeze
	funcOutp(M141->adc141+1, 0);

	conversion = (long)c1 + ((long)c2) * 256l + ((long)c3) * 65536l;

	avgConv = avgConv + conversion;

	if (tempavg++ != M141->average)
	{
	 goto Get_Conv_Again;
	}

	return avgConv/M141->average;

Get_Conv_Error:
	return -1;
}

*/
double GetConv(T_M141  *M141)
{

	static unsigned long	tempavg, c1, c2, c3;
	static unsigned long avgConv, conversion;
	double v=0;

	tempavg = 1;
	avgConv = 0;

	//select channel
	funcOutp(M141->adc141, M141->controlCode+M141->biPolar);

Get_Conv_Again:
	if (ReadStatus(M141, M141->adc141+3)==0)
	{
	 goto Get_Conv_Error;
	}

	//assert FREEZE
	funcOutp(M141->adc141+1, 1);

	c3 = (funcInp(M141->adc141+2))&0xF;
	c2 = funcInp(M141->adc141+1);
	c1 = funcInp(M141->adc141);
/*
	if ((funcInp(M141->adc141+3))&16)
	{
	 M141->digInp = 0;
	}
	else
	{
	 M141->digInp = 1;
	}
*/
	//unfreeze
	funcOutp(M141->adc141+1, 0);

	conversion = (long)c1 + ((long)c2) * 256l + ((long)c3) * 65536l;

	avgConv = avgConv + conversion;

	if (tempavg++ != M141->average)
	{
	 goto Get_Conv_Again;
	}

	if (M141->biPolar==0x80)
	{
	 v = ((double)avgConv/M141->average)*9.536743e-6 - 5;
	}
	else
	{
	 v = ((double)avgConv/M141->average)*4.768372e-6;
	}

	return v;

Get_Conv_Error:
	return -10;
}


short ReadStatus(T_M141 *M141, unsigned int status)
{
	unsigned short readStatus = 1;
	short retVal = 0;
	// create adequate delay to allow for completed conversion
	clock_t end = clock() + 500;
	while(clock() < end)
	{
		retVal = funcInp(status);
		if(retVal <= 127)
		{
			if(retVal & 16)
			{
				M141->digInp = 0;
			}
			else
			{
				M141->digInp = 1;
			}

			return readStatus;
		}
	}
	return 0;
}

short Call141(T_M141 *M141, unsigned long *clientDataBuffer)
{
	static unsigned long	numScans, dataPoint;
	static unsigned int 	firstPoint, dTime0, dTime1;
	static unsigned long avgConv, conversion;
	static int				chan, callStat, tempavg;
	static unsigned long c1, c2, c3;
	double				   eTime, dTime;

	static union
	{
		unsigned long  d;
		unsigned short i[2];
	} delayTime;

	dTime = 1/M141->hertzRate;
	delayTime.d = (M141->ichDelay*1000l);
	dTime0 = delayTime.i[0];
	dTime1 = delayTime.i[1];
	firstPoint = 0;
	numScans = 0;
	dataPoint = 0;

	//select channel
	funcOutp(M141->adc141, M141->cCode[1]+M141->biPolar);

Restart_Scan:
	eTime = (GetTickCount()*1.0)+(dTime * 1000);

Next_Scan:
	chan = 1;

Next_Channel:
	tempavg = 1;
	avgConv = 0;

More_Average:
	if (ReadStatus(M141, M141->adc141+3)==0)
	{
	 callStat = 0;
	 goto Status_Error;
	}
	//assert FREEZE
	funcOutp(M141->adc141+1, 1);

	c3 = (funcInp(M141->adc141+2))&0xF;
	c2 = funcInp(M141->adc141+1);
	c1 = funcInp(M141->adc141);

	if (M141->digInp == 1)
	{
		 callStat = 2;
		 goto Status_Error;
	}

	//unfreeze
	funcOutp(M141->adc141+1, 0);

	conversion = (long)c1 +
					((long)c2) * 256l +
					((long)c3) * 65536l;

	//if (tempavg == M141->average)
	if (tempavg == (int)M141->average)
	{
		 //select channel
		 funcOutp(M141->adc141,M141->cCode[++chan]+M141->biPolar);
	}

	avgConv = avgConv + conversion;

	if (firstPoint == 0)
	{
		 firstPoint = 1;
		 goto Restart_Scan;
	}

	//if (tempavg++ != M141->average)
	if (tempavg++ != (int)M141->average)
	{
		goto More_Average;
	}

	if (M141->numData > 1)
	{
		//do interchannel delay
		clock_t start;

		//start = clock() + ((dTime0 + (dTime1 * 0xffff)) * .001);
		start = (long)(clock() + 
				((dTime0  + (dTime1 * 0xffff )) * .001));		//8-4-98
		while(clock() < start);
	}

	clientDataBuffer[dataPoint++] = (unsigned long)(avgConv/M141->average);

	//if (chan <= M141->numData)
	if (chan <= (int)M141->numData)
	{
		goto Next_Channel;
	}

	//scan_over
	if (++numScans == M141->totScans)
	{
		callStat = 1;
		goto Status_Error;
	}

Scan_Delay:

	if (eTime > (double)GetTickCount())
	{
		goto Scan_Delay;
	}

	eTime += (dTime * 1000);

	goto Next_Scan;

Status_Error:
	//unfreeze
	funcOutp(M141->adc141+1, 0);
	M141->numCompleted = numScans;
	return callStat;
}

void SetWin32CursorOn(void)
{
	ConsoleCursorInfo.bVisible = TRUE;
	SetConsoleCursorInfo(hOutCon, &ConsoleCursorInfo);
}

void SetWin32CursorOff(void)
{
	ConsoleCursorInfo.bVisible = FALSE;
	SetConsoleCursorInfo(hOutCon, &ConsoleCursorInfo);
}


void funcOutp(UINT outPort, int outVal)
{
  short _outVal = (short)outVal;
  short _outPort = (short)outPort;
  _asm mov ax, _outVal;
  _asm mov dx, _outPort;
  __emit__(0xee);
}

short funcInp(UINT iVal)
{
  BYTE retVal = 0;
  short _iVal = (short)iVal; 
  _asm mov dx, _iVal;
  __emit__(0xec);
  _asm mov retVal, al		
  return (short)retVal;		
}


